/*
 * Portions of this file Copyright 1999-2005 University of Chicago
 * Portions of this file Copyright 1999-2005 The University of Southern California.
 *
 * This file or a portion of this file is licensed under the
 * terms of the Globus Toolkit Public License, found at
 * http://www.globus.org/toolkit/download/license.html.
 * If you redistribute this file, with or without
 * modifications, you must include this notice in the file.
 */
package org.globus.mds.webmds;

import java.io.IOException;
import java.io.PrintWriter;
import java.io.File;
import java.io.FileInputStream;
import java.util.Enumeration;
import java.util.Set;
import java.util.StringTokenizer;
import java.io.FileNotFoundException;
import javax.servlet.ServletException;
import javax.servlet.ServletContext;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.ServletConfig;
import org.globus.mds.webmds.WebmdsXmlSource;
import org.globus.mds.webmds.WebmdsUtil;
import org.globus.mds.webmds.ParameterMap;
import org.globus.mds.webmds.WebmdsConstants;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.stream.StreamResult;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerConfigurationException;
import org.xml.sax.InputSource;
import org.apache.xalan.transformer.TransformerIdentityImpl;
import org.globus.mds.webmdsConfig.WebmdsConfigType;
import org.globus.mds.webmdsConfig.WebmdsGlobalConfigType;
import org.globus.wsrf.encoding.ObjectDeserializer;
import org.globus.wsrf.encoding.DeserializationException;
import org.globus.wsrf.utils.FaultHelper;
import org.oasis.wsrf.faults.BaseFaultType;
import org.apache.axis.deployment.wsdd.WSDDNonFatalException;



/**
 * Webmds servlet -- acquires some XML and applies some XSL to it.
 * Several servlet configuration parameters control where the XML and
 * XSL come from.
 * <dl>
 * <dt><b>xmlSources</b></dt>
 * <dd>a space-separated list of <i>sourceNames</i> -- short
 * names of sources of XML.  These names will appear in other configuration
 * parameters.</dd>
 * <dt><b>info</b></dt>
 * <dd>A <i>sourceName</i> specifying where the information
 * to be displayed should come from.</dd>
 * <dt><b>xsl</b></dt>
 * <dd>A <i>sourceName</i> specifying where the xml transform
 * (that will be applied to the information) will come from.</dd>
 * <dt><b>xmlSource.</b><i>sourceName</i><b>.class</b></dt>
 * <dd> the name a class that implements the WebmdsXmlSource interface.
 * This class will be to be used to acquire XML if <i>sourceName</i>
 * is specified as an <b>info</b> or <b>xsl</b> parameter</dd>
 * <dt><b>xmlSource.</b><i>sourceName</i><b>.param.</b><i>class_specific_params</i></dt>
 * <dd>additional parameters to be passed to the xmlSource.<i>sourceName</i>.class.
 * When these parameters are passed, the <i>class_specific_params</i> names
 * will be used (in other words, the "xmlSource.<i>sourceName</i>.param." prefix
 * will be stripped off).</dd>
 * </dl>
 * @see WebmdsXmlSource
 */

public final class WebmdsServlet extends HttpServlet {
    private static final String XML_CONFIG_DIRECTORY_NAME="configDirectory";
    public static final String INFO_SOURCE_ARG_NAME="info";
    public static final String XSL_SOURCE_ARG_NAME="xsl";
    public static final String REQUEST_XSL_PARAM_PREFIX="xslParam";
    private static final String URI_XSL_PARAM="baseURI";
    private static final String INFO_SOURCE_INFO_XSL_PARAM="requestInfo";
    private static final String INFO_SOURCE_NAME_XSL_PARAM="infoSource";
    public static final String BASE_DIRECTORY_NAME="baseDirectory";
    private static final String DEFAULT_GLOBAL_CONFIG_FILE_NAME="globalconfig.xml";
    private static final String GLOBAL_CONFIG_FILE_PROPERTY_NAME="webmds.globalConfig";
    private static final String DEFAULT_CONTENT_TYPE="text/html";
    private String contentType = null;
    private ParameterMap initParameters = null;
    private WebmdsConfigEntry[] configEntryArray = null;
    private WebmdsUtil webmdsUtil = null;
    private WebmdsGlobalConfigType globalConfig = null;

    /**
     * Respond to a GET request for the content produced by
     * this servlet.
     *
     * @param request The servlet request we are processing
     * @param response The servlet response we are producing
     *
     * @exception IOException if an input/output error occurs
     * @exception ServletException if a servlet error occurs
     */
    public void doGet(HttpServletRequest request,
                      HttpServletResponse response)
	throws IOException, ServletException {

 	PrintWriter writer = response.getWriter();
	this.contentType = null;

	ParameterMap requestParameters = getRequestParameters(request);
	if (requestParameters == null) {
	    if (this.newStyleErrors()) {
		this.fail(response,
			  "No " + this.INFO_SOURCE_ARG_NAME +
			  " argument was specified to WebMDS.  This probably indicates an error in the HTML form or URL used to invoke WebMDS");
		return;
	    } else {
		throw new WebmdsException("No " + this.INFO_SOURCE_ARG_NAME +
					  " argument was specified.");
	    }
	}

	Parameter infoSourceParam =
	    (Parameter)requestParameters.get(this.INFO_SOURCE_ARG_NAME);
	if (infoSourceParam == null) {
	    if (this.newStyleErrors()) {
		this.fail(response, "No " + this.INFO_SOURCE_ARG_NAME +
			  " argument was specified to WebMDS.  This probably indicates an error in the HTML form or URL used to invoke WebMDS");
		return;
	    } else {
		throw new WebmdsException("No " + this.INFO_SOURCE_ARG_NAME +
					  " argument was specified.");
	    }
	}

	String infoSourceName;
	WebmdsXmlSource infoXmlSource;
	Source infoSource;
	String xslSourceName = null;

	if (this.newStyleErrors()) {
	    try {
		infoSourceName = infoSourceParam.getSingleValue();
		infoXmlSource = getXmlSource(infoSourceName, false);
		infoSource = getSource(infoSourceName,
				       infoXmlSource,
				       requestParameters);
		if (infoSource == null) {
		    displayResults(writer, "Your query returned null results");
		    return;
		}
 	    
		Parameter xslSourceParam =
		    (Parameter)requestParameters.get(this.XSL_SOURCE_ARG_NAME);
		if (xslSourceParam != null) {
		    xslSourceName = xslSourceParam.getSingleValue();
		}
	    } catch(WebmdsException e) {
		fail(response, e);
		return;
	    } catch(BaseFaultType e) {
		fail(response, e);
		return;
	    }
	} else {
	    infoSourceName = infoSourceParam.getSingleValue();
	    infoXmlSource = getXmlSource(infoSourceName, false);
	    infoSource = getSource(infoSourceName,
				   infoXmlSource,
				   requestParameters);

	    Parameter xslSourceParam =
		(Parameter)requestParameters.get(this.XSL_SOURCE_ARG_NAME);
	    if (xslSourceParam != null) {
		xslSourceName = xslSourceParam.getSingleValue();
	    }
	}

	Transformer transformer;
	if (xslSourceName == null) {
	    transformer = new TransformerIdentityImpl();
	    this.contentType = new String("text/xml");
	} else {
	    WebmdsXmlSource xslXmlSource;
	    Source xslSource;

	    if (this.newStyleErrors()) {
		try {
		    xslXmlSource = getXmlSource(xslSourceName, true);
		    xslSource = getSource(xslSourceName,
					  xslXmlSource,
					  requestParameters);
		    
		    Parameter contentTypeParam =
			(Parameter)this.initParameters.get(
			    WebmdsConfigEntry.getContentTypeParamName(xslSourceName));

		    if (contentTypeParam != null) {
			this.contentType = contentTypeParam.getSingleValue();
		    }
		} catch(WebmdsException e) {
		    fail(response, e);
		    return;
		} catch(BaseFaultType e) {
		    fail(response, e);
		    return;
		}
	    } else {
		xslXmlSource = getXmlSource(xslSourceName, true);
		xslSource = getSource(xslSourceName,
				      xslXmlSource,
				      requestParameters);

		Parameter contentTypeParam =
		    (Parameter)this.initParameters.get(
			WebmdsConfigEntry.getContentTypeParamName(xslSourceName));

		if (contentTypeParam != null) {
		    this.contentType = contentTypeParam.getSingleValue();
		}
	    }
	    
	    TransformerFactory transformerFactory =
		TransformerFactory.newInstance();

	    try {
		transformer = 
		    transformerFactory.newTransformer(xslSource);
	    } catch (TransformerConfigurationException e) {
		if (this.newStyleErrors()) {
		    this.fail(response, "Error initializing XSL transformation.  This usually indicates a syntax error in the specified XSL file",
			      e);
		    return;
		} else {
		    throw new WebmdsException("Error initializing xsl transformation",
					      e);
		}
	    }

	    setXslParameters(transformer, requestParameters);
	}

	transformer.setParameter(this.INFO_SOURCE_INFO_XSL_PARAM,
				 infoXmlSource.getRequestInfo());

	transformer.setParameter(this.INFO_SOURCE_NAME_XSL_PARAM,
				 infoSourceName);

	String uriParam = nonXslURIString(request, requestParameters);

	transformer.setParameter(URI_XSL_PARAM,
				 nonXslURIString(request, requestParameters));

	if (contentType == null) {
	    contentType = DEFAULT_CONTENT_TYPE;
	}
	response.setContentType(contentType);

 	try {
 	    transformer.transform(infoSource, 
 				  new StreamResult(writer));
 	} catch (TransformerException e) {
	    if (this.newStyleErrors()) {
		this.fail(response, "Error performing xsl transformation.  This may indicate badly formatted XML from the XML source or an error in the XSL file.",
			  e);
	    } else {
		throw new WebmdsException("Error performing xsl transformation",
					  e);
	    }
 	    return;

	}

    }

    /**
     * Initialize the servlet (read configuration).
     */
    public void init(ServletConfig config) throws ServletException
    {
	String configDirectoryNameString =
	    config.getInitParameter(this.XML_CONFIG_DIRECTORY_NAME);

	if (configDirectoryNameString == null) {
	    throw new WebmdsException("No " + this.XML_CONFIG_DIRECTORY_NAME +
				       " parameter has been configured.");
	}
	this.initParameters = new ParameterMap();

	// Initialize util object
	this.webmdsUtil = new WebmdsUtil();

	ServletContext context = config.getServletContext();
	String baseDirectory = context.getRealPath("/");
	if (baseDirectory != null) {
	    this.webmdsUtil.setDirectoryName(baseDirectory);
	}

	// Add config parameters
	Enumeration paramNames = config.getInitParameterNames();
	while (paramNames.hasMoreElements()) {
	    String name = (String)paramNames.nextElement();
	    if (name != null) {
		Parameter parameter =
		    Parameter.createSingleValuedParameter(
			name,
			config.getInitParameter(name));
		this.initParameters.putParameter(parameter);
	    }
	}

	// Add global config parameters
	this.readGlobalConfig();

	Parameter param = Parameter.createSingleValuedParameter(
	    WebmdsConstants.NEW_STYLE_ERROR_PARAM_NAME,
	    Boolean.toString(this.newStyleErrors()));
	this.initParameters.putParameter(param);
 
	File configDirectory =
	    this.webmdsUtil.newFile(configDirectoryNameString);
	if (configDirectory == null) {
	    throw new ServletException("Bad configuration directory '" +
				       configDirectoryNameString + "'");
	}
	File[] configFiles = configDirectory.listFiles();
	if (configFiles == null) {
	    throw new WebmdsException("Bad configuration directory '" +
				      configDirectoryNameString + "'");
	}
	this.configEntryArray = new WebmdsConfigEntry[configFiles.length];
	for (int i = 0; i < configFiles.length; i++) {
	    try {
		configEntryArray[i] = initWebmdsConfigEntry(configFiles[i]);
	    } catch (IOException e) {
		throw new WebmdsException(e);
	    }
	}
    }

    private WebmdsConfigEntry initWebmdsConfigEntry(File configFile)
	throws ServletException, IOException {
	WebmdsConfigType configType = new WebmdsConfigType();
	FileInputStream inFile = new FileInputStream(configFile);
	InputSource source = new InputSource(inFile);
	WebmdsConfigType config;
	try {
	    config = (WebmdsConfigType)ObjectDeserializer.deserialize(
		source, WebmdsConfigType.class);
	} catch (DeserializationException e) {
	    throw new WebmdsException(e);
	} catch (java.lang.NullPointerException e) {
	    throw new WebmdsConfigException("Error parsing a wsdd file.  This may mean that the servlet container (typically Tomcat) was started from the $GLOBUS_LOCATION directory, in which case restarting from a different directory will fix the problem.", e);
	}
	inFile.close();
	if ((config == null) || (config.getClassName() == null)) {
	    throw new ServletException("Error in config file " +
				       configFile.getAbsolutePath());
	}
	WebmdsConfigEntry configEntry =
	    new WebmdsConfigEntry(configFile.getName(), config);
	configEntry.init();
	return(configEntry);
    }

    private WebmdsXmlSource getXmlSource(String sourceName, boolean checkXSL)
    throws WebmdsException {
	WebmdsConfigEntry config = this.findConfigEntry(sourceName);
	if (config == null) {
	    throw new WebmdsException("No configuration information for '" +
				      sourceName + "'");
	}
	if (checkXSL) {
	    if (! config.isSuitableForXSL()) {
		throw new WebmdsException("The '" + sourceName + "' XML source has not been configured to be used as a source of XSL.  To change this, set the " + WebmdsConfigEntry.SUITABLE_FOR_XSL_NAME + " parameter to 'true' in the config entry for '" + sourceName + "' on the WebMDS server.");
	    }
	    if (this.contentType == null) {
		this.contentType = config.getContentType();
	    }
	}

	WebmdsXmlSource webmdsXmlSource;
	try {
	    webmdsXmlSource =
		(WebmdsXmlSource)config.getSourceClass().newInstance();
	} catch (java.lang.InstantiationException e) {
	    throw new WebmdsException("Error instantiating xml source class for " +
				      sourceName,
				      e);
	} catch (java.lang.IllegalAccessException e) {
	    throw new WebmdsException("Error instantiating xml source class for " +
				      sourceName,
				      e);
	}
	webmdsXmlSource.init(this.webmdsUtil);
	return(webmdsXmlSource);
    }

    private Source getSource(String sourceName,
			     WebmdsXmlSource webmdsXmlSource,
			     ParameterMap requestParams)
	throws ServletException, IOException {
	WebmdsConfigEntry config = this.findConfigEntry(sourceName);
	if (config == null) {
	    throw new WebmdsException("No configuration information for " +
				      sourceName);
	}

	ParameterMap submappedRequestParams =
	    requestParams.prefixedSubMap(
		WebmdsConfigEntry.getCustomParamPrefix(sourceName));

	ParameterMap configParams = config.getParams();

	/*
	 * hack to tell the xml source whether the user-specified-query
	 * feature is enabled.
	 */
	Parameter param = Parameter.createSingleValuedParameter(
	    WebmdsConstants.ALLOW_USER_SPECIFIED_QUERY_PARAM_NAME,
	    Boolean.toString(this.allowUserSpecifiedQuery()));
	configParams.putParameter(param);

	return(webmdsXmlSource.getXmlSource(sourceName,
					    configParams,
					    submappedRequestParams));

    }

    private WebmdsConfigEntry findConfigEntry(String sourceName) {
	for (int i = 0; i < this.configEntryArray.length; i++) {
	    if (configEntryArray[i].getName().equals(sourceName)) {
		return(configEntryArray[i]);
	    }
	}
	return(null);
    }

    private ParameterMap getRequestParameters(HttpServletRequest request)
	throws ParameterMapException {
	Enumeration paramNames = request.getParameterNames();
	if (paramNames ==  null) {
	    return(null);
	}
	ParameterMap paramMap = new ParameterMap();
	while (paramNames.hasMoreElements()) {
	    String key = (String)paramNames.nextElement();
	    if (key != null) {
		Parameter parameter =
		    new Parameter(key, request.getParameterValues(key));
		paramMap.putParameter(parameter);
	    }
	}
	if (paramMap.size() == 0) {
	    return(null);
	}
	return(paramMap);
    }

    private void setXslParameters(Transformer transformer,
				  ParameterMap paramMap)
	throws WebmdsException
    {
	if (paramMap == null) {
	    return;
	}

	ParameterMap xslParamMap =
	    paramMap.prefixedSubMap(this.REQUEST_XSL_PARAM_PREFIX);

	if (xslParamMap == null) {
	    return;
	}
	Set keySet = xslParamMap.keySet();
	if (keySet == null) {
	    return;
	}
	String[] keys = (String[])keySet.toArray(new String[0]);
	for (int i = 0; i < keys.length; i++) {
	    Parameter xslParam = (Parameter)xslParamMap.get(keys[i]);
	    transformer.setParameter(xslParam.getName(),
				     xslParam.getSingleValue());
	}
    }

    /**
     * nonXslURIString -- returns a URI that consists of the base URI
     * without the xsl-specific parameters ("xsl", "xslParam", etc.);
     */
    private String nonXslURIString(HttpServletRequest request,
				   ParameterMap paramMap)
    {
	String separator = "?";
	StringBuffer uriBuf = new StringBuffer(request.getRequestURI());
	if (paramMap == null) {
	    return(uriBuf.toString());
	}
	Set keySet = paramMap.keySet();
	if (keySet == null) {
	    return(uriBuf.toString());
	}
	String[] keys = (String[])keySet.toArray(new String[0]);
	for (int i = 0; i < keys.length; i++) {
	    Parameter param = (Parameter)paramMap.get(keys[i]);
	    if (! (param.getName().equals(XSL_SOURCE_ARG_NAME) ||
		   (param.getName().startsWith(REQUEST_XSL_PARAM_PREFIX)))) {
		String[] values = param.getValues();
		if (values != null) {
		    for (int j = 0; j < values.length; j++) {
			if (values[j] != null) {
			    uriBuf.append(separator);
			    uriBuf.append(param.getName());
			    uriBuf.append("=");
			    uriBuf.append(values[j]);
			    separator = "&";
			}
		    }
		}
	    }
	}
	return(uriBuf.toString());
    }

     private void fail(HttpServletResponse response, String introMessage, Throwable e)
 	throws IOException {
 	response.setStatus(HttpServletResponse.SC_INTERNAL_SERVER_ERROR);
 	PrintWriter out = response.getWriter();
 	String errorType = (e == null ? "WebMDS Error" : "WebMDS Exception");
 	out.println("<html>");
 	out.println("<head><title>" + errorType + "</title></head>");
 	out.println("<body>");
 	out.println("<h2>" + errorType + "</h2>");
 	
 	FaultHelper helper = null;
 	Throwable t = e;
 	String message = null;
 
 	if (introMessage != null) {
 	    out.println("<p>");
 	    out.println(introMessage);
 	    out.println("</p>");
 	}
 	
 	while (message == null && t != null) {
 	    message = t.getMessage();
 	    t = t.getCause();
 	}
 
 	String descriptions[] = null;
 	if (e != null) {
 	    if (e instanceof BaseFaultType) {
 		helper = new FaultHelper((BaseFaultType)e, true);
 		descriptions = helper.getDescription();
 		message = helper.getMessage();
 	    } else {
 		message = e.getMessage();
 	    }
 	    if (message == null && descriptions == null) {
 		Throwable cause = e.getCause();
 		if (cause instanceof BaseFaultType) {
 		    helper = new FaultHelper((BaseFaultType)cause, true);
 		    message = helper.getMessage();
 		    descriptions = helper.getDescription();
 		} else {
 		    message = cause.getMessage();
 		}
 	    }
 	}
 
 	if (descriptions != null && descriptions[0] != null) {
 	    out.println("<h3>Description</h3>");
 	    for (int i = descriptions.length; i > 0; ) {
 		out.println("<p>");
 		out.println(descriptions[--i]);
 		out.println("</p>");
 	    }
 	}
 
 	if (message != null) {
 	    out.println("<h3>Message</h3>");
 	    out.println("<p>");
 	    out.println(message);
 	    out.println("</p>");
 	}
 
 	out.println("<h3>Detail:</h3>");
 	out.println("<pre>");
 	int i = 0;
 	while (e != null && i++ < 10) {
 	    e.printStackTrace(out);
 	    if (e instanceof BaseFaultType) {
 		helper = new FaultHelper((BaseFaultType)e, true);
 		helper.printStackTrace(out);
 		break;
 	    }
 	    e = e.getCause();
 	}
 	out.println("</pre>");
 	out.println("</body>");
 	out.println("</html>");
     }
 
    private void fail(HttpServletResponse response, String message)
 	throws IOException {
 	fail(response, message, null);
    }

    private void fail(HttpServletResponse response, Throwable e)
 	throws IOException {
 	fail(response, null, e);
    }

    private void readGlobalConfig() throws WebmdsConfigException {
	String fileName =
	    System.getProperty(this.GLOBAL_CONFIG_FILE_PROPERTY_NAME);
	if (fileName == null) {
	    fileName = this.DEFAULT_GLOBAL_CONFIG_FILE_NAME;
	}
	File configFile = this.webmdsUtil.newFile(fileName);

	FileInputStream configStream;
	try {
	    configStream = new FileInputStream(configFile);
	} catch (FileNotFoundException e) {
	    // if no config file exists, stick to defaults
	    return;
	}

	try {
	    InputSource source = new InputSource(configStream);
	    this.globalConfig =
		(WebmdsGlobalConfigType)ObjectDeserializer.deserialize(
		    source, WebmdsGlobalConfigType.class);
	} catch (DeserializationException e) {
	    throw new WebmdsConfigException("Error parsing global configuration file " + configFile.getAbsolutePath(), e);
	} catch (java.lang.NullPointerException e) {
	    throw new WebmdsConfigException("Error parsing a wsdd file.  This may mean that the servlet container (typically Tomcat) was started from the $GLOBUS_LOCATION directory, in which case restarting from a different directory will fix the problem.", e);
	}
    }

    private void displayResults(PrintWriter out, String results) {
 	out.println("<html>");
 	out.println("<head><title>WebMDS Results</title></head>");
 	out.println("<body>");
 	out.println("<p>");
 	out.println(results);
 	out.println("</p>");
 	out.println("</body>");
 	out.println("</html>");
    }

    private boolean newStyleErrors() {
	/*
	 * Return the value of the oldStyleErrors config parameter.
	 * If it's not configured, return false.
	 */
	boolean result = false;
	if (this.globalConfig != null) {
	    result = this.globalConfig.isNewStyleErrors();
	}
	return(result);
    }

    private boolean allowUserSpecifiedQuery() {
	/*
	 * Return the value of the allowUserSpecifiedQuery config parameter.
	 * If it's not configured, return true.
	 */
	boolean result = false;
	if (this.globalConfig != null) {
	    result = this.globalConfig.isAllowUserSpecifiedQuery();
	}
	return(result);
    }
}

